# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import (
    List,
    Optional,
    Tuple,
    Union,
)

import forge
import numpy as np
from qctrlcommons.node.base import Node
from qctrlcommons.preconditions import (
    check_argument_hermitian,
    check_argument_integer,
    check_numeric_numpy_array,
    check_operator,
)
from scipy.sparse import spmatrix

from .documentation import Category
from .node_data import (
    Pwc,
    SparsePwc,
    Tensor,
)
from .types import TensorLike
from .utils import (
    check_argument,
    check_density_matrix_shape,
    check_lindblad_terms,
    check_oqs_hamiltonian,
    check_sample_times_with_bounds,
)


class DensityMatrixEvolutionPwc(Node):
    r"""
    Calculate the state evolution of an open system described by the GKS–Lindblad master
    equation.

    The controls that you provide to this function have to be in piecewise-constant
    format. If your controls are smooth sampleable tensor-valued functions (STFs), you
    have to discretize them with `discretize_stf` before passing them to this function.
    You may need to increase the number of segments that you choose for the
    discretization depending on the sizes of oscillations in the smooth controls.

    By default, this function computes an approximate piecewise-constant solution for
    the consideration of efficiency, with the accuracy controlled by the
    parameter `error_tolerance`. If your system is small, you can set the
    `error_tolerance` to None to obtain an exact solution.

    Note that when using the exact method, both `hamiltonian` and `lindblad_terms`
    are converted to the dense representation, regardless of their original formats.
    This means that the computation can be slow and memory intensive when applied to large systems.

    When using the approximate method, the sparse representation is used internally if
    `hamiltonian` is a `SparsePwc`, otherwise the dense representation is used.

    Parameters
    ----------
    initial_density_matrix : np.ndarray or Tensor
        A 2D array of the shape ``(D, D)`` representing the initial density matrix of
        the system, :math:`\rho_{\rm s}`. You can also pass a batch of density matrices
        and the input data shape must be ``(B, D, D)`` where ``B`` is the batch dimension.
    hamiltonian : Pwc or SparsePwc
        A piecewise-constant function representing the effective system Hamiltonian,
        :math:`H_{\rm s}(t)`, for the entire evolution duration.
    lindblad_terms : list[tuple[float, np.ndarray or Tensor or scipy.sparse.spmatrix]]
        A list of pairs, :math:`(\gamma_j, L_j)`, representing the positive decay rate
        :math:`\gamma_j` and the Lindblad operator :math:`L_j` for each coupling
        channel :math:`j`. You must provide at least one Lindblad term.
    sample_times : list or tuple or np.ndarray or None, optional
        A 1D array of length :math:`T` specifying the times :math:`\{t_i\}` at which this
        function calculates system states. Must be ordered and contain at least one element.
        Note that increasing the density of sample times does not affect the computation precision
        of this function. If omitted only the evolved density matrix at the final time of the
        system Hamiltonian is returned.
    error_tolerance : float or None, optional
        Defaults to 1e-6. This option enables an approximate method to solve the master
        equation, meaning the 2-norm of the difference between the propagated state and the exact
        solution at the final time (and at each sample time if passed) is within the error
        tolerance. Note that, if set, this value must be smaller than 1e-2 (inclusive).
        However, setting it to a too small value (for example below 1e-12) might result in slower
        computation, but would not further improve the precision, since the dominating error in
        that case is due to floating point error. You can also explicitly set this option to
        None to find the exact piecewise-constant solution. Note that using the exact solution
        can be very computationally demanding in calculations involving a large Hilbert space or
        a large number of segments.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor
        The time-evolved density matrix, with shape ``(D, D)`` or ``(T, D, D)``,
        depending on whether you provided sample times.
        If you provide a batch of initial states, the shape is ``(B, T, D, D)`` or ``(B, D, D)``.

    See Also
    --------
    Graph.jump_trajectory_evolution_pwc :
        Trajectory-based state evolution of an open quantum system.
    Graph.steady_state : Compute the steady state of open quantum system.

    Notes
    -----
    Under the Markovian approximation, the dynamics of an open quantum system can be described by
    the GKS–Lindblad master equation [1]_ [2]_

    .. math::
        \frac{{\rm d}\rho_{\rm s}(t)}{{\rm d}t} = -i [H_{\rm s}(t), \rho_{\rm s}(t)]
        + \sum_j \gamma_j {\mathcal D}[L_j] \rho_{\rm s}(t) ,

    where :math:`{\mathcal D}` is a superoperator describing the decoherent process in the
    system evolution and defined as

    .. math::
        {\mathcal D}[X]\rho := X \rho X^\dagger
            - \frac{1}{2}\left( X^\dagger X \rho + \rho X^\dagger X \right)

    for any system operator :math:`X`.

    This function uses sparse matrix multiplication when the Hamiltonian is passed as a
    `SparsePwc` and the Lindblad operators as sparse matrices. This leads to more efficient
    calculations when they involve large operators that are relatively sparse (contain mostly
    zeros). In this case, the initial density matrix is still a densely represented array or tensor.

    References
    ----------
    .. [1] `V. Gorini, A. Kossakowski, and E. C. G. Sudarshan,
            J. Math. Phys. 17, 821 (1976).
            <https://doi.org/10.1063/1.522979>`_
    .. [2] `G. Lindblad,
            Commun. Math. Phys. 48, 119 (1976).
            <https://doi.org/10.1007/BF01608499>`_

    Examples
    --------
    Simulate a trivial decay process for a single qubit described by the following master equation
    :math:`\dot{\rho} = -i[\sigma_z / 2, \, \rho] + \mathcal{D}[\sigma_-]\rho`.

    >>> duration = 20
    >>> initial_density_matrix = np.array([[0, 0], [0, 1]])
    >>> hamiltonian = graph.constant_pwc_operator(
    ...     duration=duration, operator=graph.pauli_matrix("Z") / 2
    ... )
    >>> lindblad_terms = [(1, graph.pauli_matrix("M"))]
    >>> graph.density_matrix_evolution_pwc(
    ...     initial_density_matrix, hamiltonian, lindblad_terms, name="decay"
    ... )
    <Tensor: name="decay", operation_name="density_matrix_evolution_pwc", shape=(2, 2)>
    >>> result = bo.execute_graph(graph=graph, output_node_names="decay")
    >>> result["output"]["decay"]["value"]
    array([[9.99999998e-01+0.j, 0.00000000e+00+0.j],
           [0.00000000e+00+0.j, 2.06115362e-09+0.j]])

    See more examples in the `How to simulate open system dynamics
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-simulate-open-system-dynamics>`_
    and `How to simulate large open system dynamics
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-simulate-large-open-system-dynamics>`_
    user guides.
    """

    name = "density_matrix_evolution_pwc"
    args = [
        forge.arg("initial_density_matrix", type=TensorLike),
        forge.arg("hamiltonian", type=Union[Pwc, SparsePwc]),
        forge.arg(
            "lindblad_terms",
            type=List[Tuple[float, Union[np.ndarray, Tensor, spmatrix]]],
        ),
        forge.arg(
            "sample_times", type=Optional[Union[list, tuple, np.ndarray]], default=None
        ),
        forge.arg("error_tolerance", type=Optional[float], default=1e-6),
    ]
    rtype = Tensor
    categories = [Category.LARGE_SYSTEMS, Category.TIME_EVOLUTION]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        sample_times = kwargs.get("sample_times")
        initial_density_matrix = kwargs.get("initial_density_matrix")
        hamiltonian = kwargs.get("hamiltonian")
        lindblad_terms = kwargs.get("lindblad_terms")
        error_tolerance = kwargs.get("error_tolerance")

        check_argument(
            isinstance(hamiltonian, (Pwc, SparsePwc)),
            "Hamiltonian must be a Pwc or a SparsePwc.",
            {"hamiltonian": hamiltonian},
        )
        if isinstance(hamiltonian, Pwc):
            check_argument(
                hamiltonian.batch_shape == (),
                "Hamiltonian cannot contain a batch.",
                {"hamiltonian": hamiltonian},
                extras={"hamiltonian.batch_shape": hamiltonian.batch_shape},
            )
        check_operator(initial_density_matrix, "initial_density_matrix")
        check_argument(
            not isinstance(initial_density_matrix, spmatrix),
            "Initial density matrix must not be sparse.",
            {"initial_density_matrix": initial_density_matrix},
        )

        if error_tolerance is not None:
            check_argument(
                error_tolerance <= 1e-2,
                "`error_tolerance` must not be greater than 1e-2.",
                {"error_tolerance": error_tolerance},
            )
        if sample_times is not None:
            sample_times = np.asarray(sample_times)
            check_sample_times_with_bounds(
                sample_times, "sample_times", hamiltonian, "hamiltonian"
            )

        check_density_matrix_shape(initial_density_matrix, "initial_density_matrix")
        system_dimension = initial_density_matrix.shape[-1]
        check_oqs_hamiltonian(hamiltonian, system_dimension, "initial_density_matrix")
        check_lindblad_terms(lindblad_terms, system_dimension, "initial_density_matrix")

        initial_state_shape = initial_density_matrix.shape
        if sample_times is None:
            shape = initial_state_shape
        else:
            shape = (
                initial_state_shape[:-2]
                + (len(sample_times),)
                + initial_state_shape[-2:]
            )
        return Tensor(_operation, shape=shape)


class JumpTrajectoryEvolutionPwc(Node):
    r"""
    Calculate the state evolution of an open system described by the GKS–Lindblad master
    equation using a jump-based trajectory method.

    This function calculates multiple pure-state trajectories starting from an initial pure state
    and returns the average density matrix over all trajectories.

    Note that regardless of their original formats, both `hamiltonian` and `lindblad_terms`
    are internally converted to a dense representation, so there is no computational advantage
    in using a sparse representation with this method.

    Parameters
    ----------
    initial_state_vector : np.ndarray or Tensor
        The initial state vector :math:`|\psi\rangle` as a ``(D,)`` array or Tensor,
        or batch of initial state vectors as a ``(B, D)`` array or Tensor.
    hamiltonian : Pwc or SparsePwc
        A piecewise-constant function representing the system Hamiltonian,
        :math:`H_{\rm s}(t)`, for the entire evolution duration,
        with Hilbert space dimension D.
    lindblad_terms : list[tuple[float, np.ndarray or Tensor or scipy.sparse.spmatrix]]
        A list of pairs, :math:`(\gamma_j, L_j)`, representing the positive decay rate
        :math:`\gamma_j` and the Lindblad operator :math:`L_j` for each coupling
        channel :math:`j`. You must provide at least one Lindblad term.
    max_time_step : float
        The maximum time step to use in the integration.
        Each PWC segment will be subdivided into steps that are, at most, this value.
        A smaller value for the maximum time step will more accurately sample the jump processes,
        but also lead to a slower computation.
    trajectory_count : int
        The number of quantum trajectories to run.
    sample_times : list or tuple or np.ndarray or None, optional
        A 1D array of length :math:`T` specifying the times :math:`\{t_i\}` at which this
        function calculates system states. Must be ordered and contain at least one element.
        If omitted only the evolved density matrix at the final time of the system Hamiltonian
        is returned. Note that increasing the density of sample times does not affect the
        computation precision of this function, but might slow down the calculation.
    seed : int or None, optional
        A seed for the random number generator.
        Defaults to None, in which case a random value for the seed is used.
    name : str, optional
        The name of the node.

    Returns
    -------
    Tensor
        The time-evolved density matrix, with shape ``(D, D)`` or ``(T, D, D)``,
        depending on whether you provided sample times.
        If you provide a batch of initial states, the shape is ``(B, T, D, D)`` or ``(B, D, D)``.

    See Also
    --------
    Graph.density_matrix_evolution_pwc : State evolution of an open quantum system.
    Graph.steady_state : Compute the steady state of open quantum system.

    Notes
    -----
    Under the Markovian approximation, the dynamics of an open quantum system can be described by
    the GKS–Lindblad master equation [1]_ [2]_

    .. math::
        \frac{{\rm d}\rho_{\rm s}(t)}{{\rm d}t} = -i [H_{\rm s}(t), \rho_{\rm s}(t)]
        + \sum_j \gamma_j {\mathcal D}[L_j] \rho_{\rm s}(t) ,

    where :math:`{\mathcal D}` is a superoperator describing the decoherent process in the
    system evolution and defined as

    .. math::
        {\mathcal D}[X]\rho := X \rho X^\dagger
            - \frac{1}{2}\left( X^\dagger X \rho + \rho X^\dagger X \right)

    for any system operator :math:`X`.

    This function solves the GKS–Lindblad master equation with an initial pure state
    :math:`\rho_{\rm s}(0) = |\psi\rangle\langle\psi|`, by calculating
    multiple quantum trajectories performing quantum jumps,
    :math:`|\tilde\psi_k(t)\rangle`, and averaging the result:

    .. math::
        \rho(t) = \frac{1}{M} \sum_{k=1}^M |\tilde\psi_k(t) \rangle\langle \tilde\psi_k(t) | .

    References
    ----------
    .. [1] `V. Gorini, A. Kossakowski, and E. C. G. Sudarshan,
            J. Math. Phys. 17, 821 (1976).
            <https://doi.org/10.1063/1.522979>`_
    .. [2] `G. Lindblad,
            Commun. Math. Phys. 48, 119 (1976).
            <https://doi.org/10.1007/BF01608499>`_
    """
    name = "jump_trajectory_evolution_pwc"
    args = [
        forge.arg("initial_state_vector", type=TensorLike),
        forge.arg("hamiltonian", type=Union[Pwc, SparsePwc]),
        forge.arg(
            "lindblad_terms",
            type=List[Tuple[float, Union[np.ndarray, Tensor, spmatrix]]],
        ),
        forge.arg("max_time_step", type=float),
        forge.arg("trajectory_count", type=int),
        forge.arg(
            "sample_times", type=Optional[Union[list, tuple, np.ndarray]], default=None
        ),
        forge.arg("seed", type=Optional[int], default=None),
    ]
    rtype = Tensor
    categories = [Category.LARGE_SYSTEMS, Category.TIME_EVOLUTION]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        initial_state_vector = kwargs.get("initial_state_vector")
        hamiltonian = kwargs.get("hamiltonian")
        lindblad_terms = kwargs.get("lindblad_terms")
        max_time_step = kwargs.get("max_time_step")
        trajectory_count = kwargs.get("trajectory_count")
        sample_times = kwargs.get("sample_times")
        seed = kwargs.get("seed")

        check_argument(
            len(initial_state_vector.shape) in (1, 2),
            "The initial state vector must be 1D or 2D.",
            {"initial_state_vector": initial_state_vector},
        )
        system_dimension = initial_state_vector.shape[-1]

        check_argument(
            isinstance(hamiltonian, (Pwc, SparsePwc)),
            "Hamiltonian must be a Pwc or a SparsePwc.",
            {"hamiltonian": hamiltonian},
        )
        if isinstance(hamiltonian, Pwc):
            check_argument(
                hamiltonian.batch_shape == (),
                "Hamiltonian cannot contain a batch.",
                {"hamiltonian": hamiltonian},
                extras={"hamiltonian.batch_shape": hamiltonian.batch_shape},
            )
        if sample_times is not None:
            check_sample_times_with_bounds(
                sample_times, "sample_times", hamiltonian, "hamiltonian"
            )
        check_oqs_hamiltonian(hamiltonian, system_dimension, "initial_state_vector")
        check_lindblad_terms(lindblad_terms, system_dimension, "initial_state_vector")
        check_argument(
            max_time_step > 0.0,
            "The maximum time step must be positive.",
            {"max_time_step": max_time_step},
        )
        check_argument(
            trajectory_count > 0,
            "The number of trajectories must be positive.",
            {"trajectory_count": trajectory_count},
        )
        if seed is not None:
            check_argument_integer(seed, "seed")

        shape = (system_dimension, system_dimension)
        if sample_times is not None:
            shape = (len(sample_times),) + shape
        if len(initial_state_vector.shape) != 1:
            shape = (initial_state_vector.shape[0],) + shape
        return Tensor(_operation, shape=shape)


class SteadyState(Node):
    r"""
    Find the steady state of a time-independent open quantum system.

    The Hamiltonian and Lindblad operators that you provide have to give
    rise to a unique steady state.

    Parameters
    ----------
    hamiltonian : Tensor or spmatrix
        A 2D array of shape ``(D, D)`` representing the time-independent
        Hamiltonian of the system, :math:`H_{\rm s}`.
    lindblad_terms : list[tuple[float, np.ndarray or Tensor or scipy.sparse.spmatrix]]
        A list of pairs, :math:`(\gamma_j, L_j)`, representing the positive decay rate
        :math:`\gamma_j` and the Lindblad operator :math:`L_j` for each coupling
        channel :math:`j`. You must provide at least one Lindblad term.
    method : str, optional
        The method used to find the steady state.
        Must be one of "QR", "EIGEN_SPARSE", or "EIGEN_DENSE".
        The "QR" method obtains the steady state through a QR decomposition and is suitable
        for small quantum systems with dense representation.
        The "EIGEN_SPARSE" and "EIGEN_DENSE"  methods find the steady state as the eigenvector
        whose eigenvalue is closest to zero, using either a sparse or a dense representation
        of the generator.
        Defaults to "QR".
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor
        The density matrix representing the steady state of the system.

    Warnings
    --------
    This function currently does not support calculating the gradient with respect to its inputs.
    Therefore, it cannot be used in a graph for a `run_optimization` or
    `run_stochastic_optimization` call, which will raise a `RuntimeError`.
    Please use gradient-free optimization if you want to perform an optimization task with this
    function. You can learn more about it in the
    `How to optimize controls using gradient-free optimization
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-using-gradient-free-optimization>`_
    user guide.

    See Also
    --------
    Graph.density_matrix_evolution_pwc : State evolution of an open quantum system.
    Graph.jump_trajectory_evolution_pwc :
        Trajectory-based state evolution of an open quantum system.

    Notes
    -----
    Under the Markovian approximation, the dynamics of an open quantum system can be described by
    the GKS–Lindblad master equation [1]_ [2]_,

    .. math::
        \frac{{\rm d}\rho_{\rm s}(t)}{{\rm d}t} = {\mathcal L} (\rho_{\rm s}(t)) ,

    where the Lindblad superoperator :math:`{\mathcal L}` is defined as

    .. math::
        {\mathcal L} (\rho_{\rm s}(t)) = -i [H_{\rm s}(t), \rho_{\rm s}(t)]
        + \sum_j \gamma_j {\mathcal D}[L_j] \rho_{\rm s}(t) ,

    where :math:`{\mathcal D}` is a superoperator describing the decoherent process in the
    system evolution and defined as

    .. math::
        {\mathcal D}[X]\rho := X \rho X^\dagger
            - \frac{1}{2}\left( X^\dagger X \rho + \rho X^\dagger X \right)

    for any system operator :math:`X`.

    This function computes the steady state of :math:`{\mathcal L}` by solving

    .. math:: \frac{{\rm d}\rho_{\rm s}(t)}{{\rm d}t} = 0 .

    The function assumes that :math:`H_{\rm s}` is time independent
    and that the dynamics generated by :math:`{\mathcal L}`
    give rise to a unique steady state. That is, the generated quantum dynamical map
    has to be ergodic [3]_.

    References
    ----------
    .. [1] `V. Gorini, A. Kossakowski, and E. C. G. Sudarshan,
            J. Math. Phys. 17, 821 (1976).
            <https://doi.org/10.1063/1.522979>`_
    .. [2] `G. Lindblad,
            Commun. Math. Phys. 48, 119 (1976).
            <https://doi.org/10.1007/BF01608499>`_
    .. [3] `D. Burgarth, G. Chiribella, V. Giovannetti, P. Perinotti, and K. Yuasa,
            New J. Phys. 15 073045 (2013).
            <https://doi.org/10.1088/1367-2630/15/7/073045>`_

    Examples
    --------
    Compute the steady state of the single qubit open system dynamics according to the
    Hamiltonian :math:`H=\omega\sigma_z` and the single Lindblad operator :math:`L=\sigma_-`.

    >>> omega = 0.8
    >>> gamma = 0.5
    >>> hamiltonian = omega * graph.pauli_matrix("Z")
    >>> lindblad_terms = [(gamma, graph.pauli_matrix("M"))]
    >>> graph.steady_state(hamiltonian, lindblad_terms, name="steady_state")
    <Tensor: name="steady_state", operation_name="steady_state", shape=(2, 2)>
    >>> result = bo.execute_graph(graph=graph, output_node_names="steady_state")
    >>> result["output"]["steady_state"]["value"]
    array([[1.+0.j 0.-0.j]
           [0.-0.j 0.-0.j]])
    """

    name = "steady_state"
    args = [
        forge.arg("hamiltonian", type=Union[Tensor, spmatrix, np.ndarray]),
        forge.arg(
            "lindblad_terms",
            type=List[Tuple[float, Union[Tensor, spmatrix, np.ndarray]]],
        ),
        forge.arg("method", type=str, default="QR"),
    ]
    rtype = Tensor
    categories = [Category.LARGE_SYSTEMS, Category.TIME_EVOLUTION]
    supports_gradient = False

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        hamiltonian = kwargs.get("hamiltonian")
        lindblad_terms = kwargs.get("lindblad_terms")
        method = kwargs.get("method")

        check_argument(
            isinstance(hamiltonian, (Tensor, spmatrix, np.ndarray)),
            "Hamiltonian must be a Tensor, spmatrix or NumPy array.",
            {"hamiltonian": hamiltonian},
        )
        if isinstance(hamiltonian, spmatrix):
            check_argument_hermitian(hamiltonian, "hamiltonian")
        elif isinstance(hamiltonian, Tensor):
            check_operator(hamiltonian, "hamiltonian")
        elif isinstance(hamiltonian, np.ndarray):
            check_numeric_numpy_array(hamiltonian, "hamiltonian")
            check_argument_hermitian(hamiltonian, "hamiltonian")

        check_lindblad_terms(lindblad_terms, hamiltonian.shape[-1], "hamiltonian")

        if method is not None:
            check_argument(
                isinstance(method, str),
                "The method must be a string.",
                {"method": method},
            )
            check_argument(
                method in ["QR", "EIGEN_DENSE", "EIGEN_SPARSE"],
                'The method must be "QR", "EIGEN_DENSE", or "EIGEN_SPARSE".',
                {"method": method},
            )

        return Tensor(_operation, shape=hamiltonian.shape)



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from functools import wraps
from typing import (
    List,
    Optional,
    Union,
)

import forge
import numpy as np
from qctrlcommons.node.base import Node
from qctrlcommons.preconditions import (
    check_argument,
    check_argument_iterable,
    check_numeric_numpy_array,
)

from .documentation import Category
from .namespace import create_namespace_doc
from .node_data import (
    Pwc,
    Tensor,
)
from .types import TensorLike
from .utils import (
    check_sample_times_with_bounds,
    validate_ms_shapes,
    validate_shape,
)


class MsPhases(Node):
    r"""
    Calculate the relative phases for all pairs of ions described by a Mølmer–Sørensen-type
    interaction when single-tone individually-addressed laser beams are used.

    Use this function to calculate the acquired phases for all ion pairs
    at the final time of the drives, or at the sample times that you provide.

    Parameters
    ----------
    drives : list[Pwc(1D, complex)]
        A list of piecewise-constant drives :math:`\{\gamma_j\}`.
        The number of drives must be the same as the number of ions :math:`N`.
        Drive values must be one-dimensional arrays and in rad/s. Drive durations must be
        in seconds. All drives must have the same total duration, but can have different numbers
        of segments. This list must contain at least two elements. In other words,
        your system must contain at least two ions.
    lamb_dicke_parameters : np.ndarray
        A 3D array of parameters :math:`\{\eta_{jkl}\}` specifying the laser-ion coupling strength.
        Its shape must be ``(3, N, N)`` where the first dimension :math:`j` indicates the axis,
        the second dimension :math:`k` indicates the collective
        mode number, and the third dimension :math:`l` indicates the ion.
    relative_detunings : np.ndarray
        A 2D array :math:`\{\delta_{jk} = \nu_{jk} - \delta\}` specifying the difference (in Hz)
        between each motional mode frequency :math:`\nu_{jk}` and the laser detuning :math:`\delta`
        (the detuning from the qubit transition frequency :math:`\omega_0`).
        Its shape must be ``(3, N)`` where the first dimension :math:`j` indicates the axis and the
        second dimension :math:`k` indicates the collective mode number.
    sample_times : list or tuple or np.ndarray or None, optional
        A 1D array of length :math:`T` specifying the times :math:`\{t_i\}` (in seconds) at which
        this function calculates the relative phases.
        If you omit it, this function calculates the phases only at the final time of the drives.
        If provided, it must be ordered and contain at least one element.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor(real)
        Acquired phases :math:`\{\phi_{jk}(t_i) + \phi_{kj}(t_i)\}` for all ion pairs.
        If you provide `sample_times`, the shape of the returned value is ``(T, N, N)``,
        where the first dimension indicates the time; the second and the third dimensions
        indicate the ion. Otherwise, the shape is ``(N, N)`` where both dimensions indicate
        the ion. The relative phases are stored as a strictly lower triangular matrix. See
        the notes part for details.

    See Also
    --------
    :func:`Graph.ions.ms_infidelity <ions.ms_infidelity>`
        Final operational infidelity of a Mølmer–Sørensen gate.
    :func:`Graph.ms_phases_multitone <ions.ms_phases_multitone>`
        Corresponding operation for a global multitone beam.
    :func:`boulderopal.ions.obtain_ion_chain_properties`
        Function to calculate the properties of an ion chain.

    Notes
    -----
    The internal and motional Hamiltonian of :math:`N` ions is

    .. math::
        H_0 = \sum_{p = 1}^{3N} \hbar\nu_p \left(a_p^\dagger a_p + \frac{1}{2}\right)
            + \sum_{j = 1}^N \frac{\hbar \omega_0}{2} \sigma_{z,j} ,

    where the axis dimension and collective mode dimension are combined into a single index
    :math:`p` for simplicity, :math:`a_p` is the annihilation operator for the mode :math:`p`,
    and :math:`\sigma_{z,j}` is the Pauli :math:`Z` operator for the ion :math:`j`.
    The interaction Hamiltonian for Mølmer–Sørensen-type
    operations in the rotating frame with respect to :math:`H_0` is:

    .. math::
        H_I(t) = i\hbar\sum_{j = 1}^N \sigma_{x, j} \sum_{p = 1}^{3N} (-\beta_{pj}^*(t)a_p +
                \beta_{pj}(t) a_p^\dagger) ,

    where :math:`\sigma_{x, j}` is the Pauli :math:`X` operator for the ion :math:`j` and
    :math:`\beta_{pj}(t) = \eta_{pj} \frac{\gamma_j(t)}{2} e^{i\delta_p t}`,
    indicating the coupling of the ion :math:`j` to the motional mode :math:`p`.
    Note that :math:`\delta_p` is the angular relative detuning, which corresponds to
    :math:`2 \pi` times the `relative_detuning` parameter that you passed.
    The corresponding unitary operation is given by [1]_

    .. math::
        U(t) = \exp\left[ \sum_{j=1}^N \sigma_{x, j} B_j(t)
                + i\sum_{j=1}^N\sum_{k=1}^{j - 1} (\phi_{jk}(t) + \phi_{kj}(t))
                \sigma_{x, j} \sigma_{x, k} \right] ,

    where

    .. math::
        B_j(t) &\equiv \sum_{p = 1}^{3N}  \left(\eta_{pj}\alpha_{pj}(t)a_p^\dagger
             - \eta_{pj}^{\ast}\alpha_{pj}^\ast(t)a_p \right) ,

        \phi_{jk}(t) &\equiv \mathrm{Im} \left[ \sum_{p=1}^{3N} \int_{0}^{t} d \tau_1
            \int_{0}^{\tau_1} d \tau_2 \beta_{pj}(\tau_1)\beta_{pk}^{\ast}(\tau_2) \right] ,

    and

    .. math::
        \alpha_{pj}(t) = \int_0^t d\tau \frac{\gamma_j(\tau)}{2} e^{i \delta_p \tau} .

    This function calculates the relative phases
    :math:`\Phi_{jk}(t_i) = \phi_{jk}(t_i) + \phi_{kj}(t_i)`
    for all ions pairs at time :math:`t_i` and stores the result in the form of a strictly
    lower triangular matrix. That is, the :math:`jk`-th element of
    that matrix records the relative phase between the ion :math:`j` and :math:`k`,
    while the elements with indices :math:`j \leq k` are zeros.

    References
    ----------
    .. [1] `C. D. B. Bentley, H. Ball, M. J. Biercuk, A. R. R. Carvalho,
            M. R. Hush, and H. J. Slatyer,
            Adv. Quantum Technol. 3, 2000044 (2020).
            <https://doi.org/10.1002/qute.202000044>`_

    Examples
    --------
    Refer to the `How to optimize error-robust Mølmer–Sørensen gates for trapped ions
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-error-robust-molmer-
    sorensen-gates-for-trapped-ions>`_ user guide to find how to use this and related nodes.
    """

    name = "ms_phases"
    args = [
        forge.arg("drives", type=List[Pwc]),
        forge.arg("lamb_dicke_parameters", type=np.ndarray),
        forge.arg("relative_detunings", type=np.ndarray),
        forge.arg(
            "sample_times", type=Optional[Union[list, tuple, np.ndarray]], default=None
        ),
    ]
    rtype = Tensor
    categories = [Category.MOLMER_SORENSEN]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        drives = kwargs.get("drives")
        sample_times = kwargs.get("sample_times")
        lamb_dicke_parameters = kwargs.get("lamb_dicke_parameters")
        relative_detunings = kwargs.get("relative_detunings")
        check_argument_iterable(drives, "drives")
        check_argument(
            len(drives) > 1, "At least two drives must be provided.", {"drives": drives}
        )
        check_argument(
            all(isinstance(drive, Pwc) for drive in drives),
            "Each of the drives must be a Pwc.",
            {"drives": drives},
        )
        check_argument(
            all(len(drive.values.shape) == 1 for drive in drives),
            "The value of each drive must be 1D.",
            {"drives": drives},
        )
        ion_count = len(drives)
        shape = (ion_count, ion_count)
        validate_ms_shapes(
            ion_count=ion_count,
            ld_values=lamb_dicke_parameters,
            ld_name="lamb_dicke_parameters",
            rd_values=relative_detunings,
            rd_name="relative_detunings",
        )
        if sample_times is not None:
            sample_times = np.asarray(sample_times)
            check_sample_times_with_bounds(
                sample_times, "sample_times", drives[0], "drives[0]"
            )
            time_count = len(sample_times)
            shape = (time_count,) + tuple(shape)
        return Tensor(_operation, shape=shape)


class MsPhasesMultitone(Node):
    r"""
    Calculate the relative phases for all pairs of ions described by a
    Mølmer–Sørensen-type interaction where the ions are being addressed by
    a multitone global beam.

    Use this function to calculate the acquired phases for all ion pairs
    at the final time of the drives, or at the sample times that you provide.

    Parameters
    ----------
    drives : list[Pwc(1D, complex)]
        A list of piecewise-constant drives :math:`\{\gamma_\xi\}` for
        each of the tones of the global beam.
        The number of drives must be the same as the number of tones of
        the global beam, :math:`M`.
        Drive values must be one-dimensional arrays and in rad/s. Drive durations must be
        in seconds. All drives must have the same total duration, but can have different numbers
        of segments.
    lamb_dicke_parameters : np.ndarray
        A 4D array of parameters :math:`\{\eta_{\xi jkl}\}` specifying
        the laser-ion coupling strength. Its shape must be ``(M, 3, N, N)``
        where the first dimension :math:`\xi` indicates the tone of
        the global beam, the second dimension :math:`j` indicates the axis,
        the third dimension :math:`k` indicates the collective
        mode number, and the fourth dimension :math:`l` indicates the ion.
    relative_detunings : np.ndarray
        A 3D array :math:`\{\delta_{\xi jk} = \nu_{jk} - \delta_\xi \}`
        specifying the difference (in Hz) between each motional mode frequency
        :math:`\nu_{jk}` and the laser detunings for each tone,
        :math:`\delta_\xi` (the detuning from the qubit transition frequency
        :math:`\omega_0`). Its shape must be ``(M, 3, N)`` where the first
        dimension :math:`\xi` indicates the tone of the global beam,
        the second dimension :math:`j` indicates the axis and the third
        dimension :math:`k` indicates the collective mode number.
    sample_times : list or tuple or np.ndarray or None, optional
        A 1D array of length :math:`T` specifying the times :math:`\{t_i\}` (in seconds) at which
        this function calculates the relative phases.
        If you omit it, this function calculates the phases only at the final time of the drives.
        If provided, it must be ordered and contain at least one element.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor(real)
        Acquired phases :math:`\{\phi_{jk}(t_i) + \phi_{kj}(t_i)\}` for all ion pairs.
        If you provide `sample_times`, the shape of the returned value is ``(T, N, N)``,
        where the first dimension indicates the time; the second and the third dimensions
        indicate the ion. Otherwise, the shape is ``(N, N)`` where both dimensions indicate
        the ion. The relative phases are stored as a strictly lower triangular matrix. See
        the notes part for details.

    See Also
    --------
    :func:`Graph.ions.ms_infidelity <ions.ms_infidelity>`
        Final operational infidelity of a Mølmer–Sørensen gate.
    :func:`Graph.ms_phases <ions.ms_phases>`
        Corresponding operation for single-tone individually-addressed beams.
    :func:`boulderopal.ions.obtain_ion_chain_properties`
        Function to calculate the properties of an ion chain.

    Notes
    -----
    The interaction Hamiltonian for Mølmer–Sørensen-type
    operations in the rotating frame for a multitone global beam is:

    .. math::
        H_I(t) = i\hbar\sum_{j = 1}^N \sigma_{x, j} \sum_{p = 1}^{3N} \sum_{\xi = 1}^M
                 (-\beta_{\xi pj}^*(t)a_p + \beta_{\xi pj}(t) a_p^\dagger) ,

    where :math:`\sigma_{x, j}` is the Pauli :math:`X` operator for the ion :math:`j` and
    :math:`\beta_{\xi pj}(t) = \eta_{\xi pj} \frac{\gamma_\xi(t)}{2} e^{i\delta_{\xi p} t}`,
    indicating the coupling of the ion :math:`j` to the motional mode :math:`p`.
    Note that :math:`\delta_p` is the angular relative detuning, which corresponds to
    :math:`2 \pi` times the `relative_detuning` parameter that you passed.
    The corresponding unitary operation is given by:

    .. math::
        U(t) = \exp\left[ \sum_{j=1}^N \sigma_{x, j} B_j(t)
                + i\sum_{j=1}^N \sum_{k=1}^{j - 1} (\phi_{jk}(t) + \phi_{kj}(t))
                \sigma_{x, j} \sigma_{x, k} \right] ,

    where

    .. math::
        B_j(t) &\equiv \sum_{p = 1}^{3N} \sum_{\xi = 1}^M
             \left(\eta_{\xi pj}\alpha_{\xi pj}(t)a_p^\dagger
             - \eta_{\xi pj}^{\ast}\alpha_{\xi pj}^\ast(t)a_p \right) ,

        \phi_{jk}(t) &\equiv \mathrm{Im} \left[ \sum_{p=1}^{3N} \sum_{\xi_1 = 1}^M
            \sum_{\xi_2 = 1}^M \int_{0}^{t} d \tau_1 \int_{0}^{\tau_1} d \tau_2
            \beta_{\xi_1 pj}(\tau_1)\beta_{\xi_2 pk}^{\ast}(\tau_2) \right] ,

    with :math:`\alpha_{\xi pj}(t) = \int_0^t d\tau \frac{\gamma_\xi(\tau)}{2}
    e^{i \delta_{\xi p} \tau}`.

    This function calculates the relative phases :math:`\phi_{jk}(t_i) + \phi_{kj}(t_i)`
    for all ions pairs at time :math:`t_i` and stores the result in the form of a strictly
    lower triangular matrix. That is, the :math:`jk`-th element of
    that matrix records the relative phase between the ion :math:`j` and :math:`k`,
    while the elements with indices :math:`j \leq k` are zeros.
    """

    name = "ms_phases_multitone"
    args = [
        forge.arg("drives", type=List[Pwc]),
        forge.arg("lamb_dicke_parameters", type=np.ndarray),
        forge.arg("relative_detunings", type=np.ndarray),
        forge.arg(
            "sample_times", type=Optional[Union[list, tuple, np.ndarray]], default=None
        ),
    ]
    rtype = Tensor
    categories = [Category.MOLMER_SORENSEN]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        drives = kwargs.get("drives")
        sample_times = kwargs.get("sample_times")
        lamb_dicke_parameters = kwargs.get("lamb_dicke_parameters")
        relative_detunings = kwargs.get("relative_detunings")
        check_argument_iterable(drives, "drives")
        check_argument(
            len(drives) > 0, "At least one drive must be provided.", {"drives": drives}
        )
        check_argument(
            all(isinstance(drive, Pwc) for drive in drives),
            "Each of the drives must be a Pwc.",
            {"drives": drives},
        )
        check_argument(
            all(len(drive.values.shape) == 1 for drive in drives),
            "The value of each drive must be 1D.",
            {"drives": drives},
        )

        check_numeric_numpy_array(lamb_dicke_parameters, "lamb_dicke_parameters")
        check_numeric_numpy_array(relative_detunings, "relative_detunings")

        ld_shape = validate_shape(lamb_dicke_parameters, "lamb_dicke_parameters")
        rd_shape = validate_shape(relative_detunings, "relative_detunings")

        tone_count = len(drives)
        ion_count = ld_shape[-1]
        check_argument(
            ld_shape == (tone_count, 3, ion_count, ion_count),
            "The Lamb-Dicke parameters must have shape (tone_count, 3, ion_count, ion_count).",
            {"lamb_dicke_parameters": lamb_dicke_parameters},
            extras={
                "ion_count": ion_count,
                "tone_count": tone_count,
                "lamb_dicke_parameters.shape": ld_shape,
            },
        )
        check_argument(
            rd_shape == (tone_count, 3, ion_count),
            "The relative detunings must have shape (tone_count, 3, ion_count).",
            {"relative_detunings": relative_detunings},
            extras={
                "ion_count": ion_count,
                "tone_count": tone_count,
                "relative_detunings.shape": rd_shape,
            },
        )

        shape = (ion_count, ion_count)
        if sample_times is not None:
            sample_times = np.asarray(sample_times)
            check_sample_times_with_bounds(
                sample_times, "sample_times", drives[0], "drives[0]"
            )
            time_count = len(sample_times)
            shape = (time_count,) + shape

        return Tensor(_operation, shape=shape)


class MsDisplacements(Node):
    r"""
    Calculate the displacements for each mode and ion combination where ions are described by
    a Mølmer–Sørensen-type interaction.

    Use this function to calculate the displacements for each ion and each mode
    at the final time of the drives, or at the sample times that you provide.

    Parameters
    ----------
    drives : list[Pwc(1D, complex)]
        A list of piecewise-constant drives :math:`\{\gamma_j\}`.
        The number of drives must be the same as the number of ions :math:`N`.
        Drive values must be one-dimensional arrays and in rad/s.
        Drive durations must be in seconds.
        All drives must have the same total duration, but can have different numbers of segments.
    lamb_dicke_parameters : np.ndarray
        A 3D array of parameters :math:`\{\eta_{jkl}\}` specifying the laser-ion coupling strength.
        Its shape must be ``(3, N, N)`` where the first dimension :math:`j` indicates the axis,
        the second dimension :math:`k` indicates the collective
        mode number, and the third dimension :math:`l` indicates the ion.
    relative_detunings : np.ndarray
        A 2D array :math:`\{\delta_{jk} = \nu_{jk} - \delta\}` specifying the difference (in Hz)
        between each motional mode frequency :math:`\nu_{jk}` and the laser detuning :math:`\delta`
        (the detuning from the qubit transition frequency :math:`\omega_0`).
        Its shape must be ``(3, N)`` where the first dimension :math:`j` indicates the axis and the
        second dimension :math:`k` indicates the collective mode number.
    sample_times : list or tuple or np.ndarray or None, optional
        A 1D array of length :math:`T` specifying the times :math:`\{t_i\}` (in seconds) at which
        this function calculates the displacements.
        If you omit it, this function calculates the phases only at the final time of the drives.
        If provided, it must be ordered and contain at least one element.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor(complex)
        Displacements :math:`\{\eta_{pj}\alpha_{pj}(t_i)\}` for all mode-ion combinations.
        If you provide `sample_times`, the shape of the returned value is ``(T, 3, N, N)``
        where the first dimension indicates the time, the second dimension indicates the axis,
        the third dimension indicates the collective mode number, and the last dimension indicates
        the ion. Otherwise, the shape is ``(3, N, N)`` with the outer time dimension removed.

    See Also
    --------
    :func:`Graph.ions.ms_infidelity <ions.ms_infidelity>`
        Final operational infidelity of a Mølmer–Sørensen gate.
    :func:`boulderopal.ions.obtain_ion_chain_properties`
        Function to calculate the properties of an ion chain.

    Notes
    -----
    The internal and motional Hamiltonian of :math:`N` ions is

    .. math::
        H_0 = \sum_{p = 1}^{3N} \hbar\nu_p \left(a_p^\dagger a_p + \frac{1}{2}\right)
            + \sum_{j = 1}^N \frac{\hbar \omega_0}{2} \sigma_{z,j} ,

    where the axis dimension and collective mode dimension are combined into a single index
    :math:`p` for simplicity, :math:`a_p` is the annihilation operator for the mode :math:`p`,
    and :math:`\sigma_{z,j}` is the Pauli :math:`Z` operator for the ion :math:`j`.
    The interaction Hamiltonian for Mølmer–Sørensen-type
    operations in the rotating frame with respect to :math:`H_0` is:

    .. math::
        H_I(t) = i\hbar\sum_{j = 1}^N \sigma_{x, j} \sum_{p = 1}^{3N} (-\beta_{pj}^*(t)a_p +
                \beta_{pj}(t) a_p^\dagger) ,

    where :math:`\sigma_{x, j}` is the Pauli :math:`X` operator for the ion :math:`j` and
    :math:`\beta_{pj}(t) = \eta_{pj} \frac{\gamma_j(t)}{2} e^{i\delta_p t}`,
    indicating the coupling of the ion :math:`j` to the motional mode :math:`p`.
    Note that :math:`\delta_p` is the angular relative detuning, which corresponds to
    :math:`2 \pi` times the `relative_detuning` parameter that you passed.
    The corresponding unitary operation is given by [1]_

    .. math::
        U(t) = \exp\left[ \sum_{j=1}^N \sigma_{x, j} B_j(t)
                + i\sum_{j=1}^N\sum_{k=1}^{j - 1} (\phi_{jk}(t) + \phi_{kj}(t))
                \sigma_{x, j} \sigma_{x, k} \right] ,

    where

    .. math::
        B_j(t) &\equiv \sum_{p = 1}^{3N}  \left(\eta_{pj}\alpha_{pj}(t)a_p^\dagger
             - \eta_{pj}^{\ast}\alpha_{pj}^\ast(t)a_p \right) ,

        \phi_{jk}(t) &\equiv \mathrm{Im} \left[ \sum_{p=1}^{3N} \int_{0}^{t} d \tau_1
            \int_{0}^{\tau_1} d \tau_2 \beta_{pj}(\tau_1)\beta_{pk}^{\ast}(\tau_2) \right] ,

    and

    .. math::
        \alpha_{pj}(t) = \int_0^t d\tau \frac{\gamma_j(\tau)}{2} e^{i \delta_p \tau} .

    This function calculates, for each time :math:`t_i`, the contribution to the displacement of
    mode :math:`p` from the ion :math:`j`, namely :math:`\eta_{pj}\alpha_{pj}(t_i)`. You can
    calculate the state-dependent displacement for the mode :math:`p` by summing over the
    contributions from all ions. That is, using the displacement superoperator
    :math:`\mathcal{D}_p`, the displacement in phase space for mode :math:`p` at time
    :math:`t_i` is:

    .. math::
        \mathcal{D}_p \left(\sum_{j = 1}^N \sigma_{x, j}\eta_{pj}\alpha_{pj}(t_i) \right) .

    References
    ----------
    .. [1] `C. D. B. Bentley, H. Ball, M. J. Biercuk, A. R. R. Carvalho,
            M. R. Hush, and H. J. Slatyer,
            Adv. Quantum Technol. 3, 2000044 (2020).
            <https://doi.org/10.1002/qute.202000044>`_

    Examples
    --------
    Refer to the `How to optimize error-robust Mølmer–Sørensen gates for trapped ions
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-error-robust-molmer-sorensen
    -gates-for-trapped-ions>`_ user guide to find how to use this and related nodes.
    """

    name = "ms_displacements"
    args = [
        forge.arg("drives", type=List[Pwc]),
        forge.arg("lamb_dicke_parameters", type=np.ndarray),
        forge.arg("relative_detunings", type=np.ndarray),
        forge.arg(
            "sample_times", type=Optional[Union[list, tuple, np.ndarray]], default=None
        ),
    ]
    rtype = Tensor
    categories = [Category.MOLMER_SORENSEN]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        drives = kwargs.get("drives")
        sample_times = kwargs.get("sample_times")
        lamb_dicke_parameters = kwargs.get("lamb_dicke_parameters")
        relative_detunings = kwargs.get("relative_detunings")
        check_argument_iterable(drives, "drives")
        check_argument(
            len(drives) > 0, "At least one drive must be provided.", {"drives": drives}
        )
        check_argument(
            all(isinstance(drive, Pwc) for drive in drives),
            "Each of the drives must be a Pwc.",
            {"drives": drives},
        )
        check_argument(
            all(len(drive.values.shape) == 1 for drive in drives),
            "The value of each drive must be 1D.",
            {"drives": drives},
        )
        ion_count = len(drives)
        shape = (3, ion_count, ion_count)
        validate_ms_shapes(
            ion_count=ion_count,
            ld_values=lamb_dicke_parameters,
            ld_name="lamb_dicke_parameters",
            rd_values=relative_detunings,
            rd_name="relative_detunings",
        )
        if sample_times is not None:
            sample_times = np.asarray(sample_times)
            check_sample_times_with_bounds(
                sample_times, "sample_times", drives[0], "drives[0]"
            )
            time_count = len(sample_times)
            shape = (time_count,) + tuple(shape)
        return Tensor(_operation, shape=shape)


class MsInfidelity(Node):
    r"""
    Calculate the final operational infidelity of the Mølmer–Sørensen gate.

    This function calculates the operational infidelity with respect to the target phases
    that you specify in the `target_phases` array. It can use the tensors returned from
    :func:`ms_phases` and :func:`ms_displacements` to calculate the infidelity tensor.

    Parameters
    ----------
    phases : np.ndarray(real) or Tensor(real)
        Acquired phases :math:`\{\Phi_{kl}\}` for all ion pairs with shape ``(N, N)`` without time
        samples or ``(T, N, N)``, where ``T`` is the number of samples and ``N`` is the
        number of ions. For each sample the `phases` array must be a strictly lower
        triangular matrix.
    displacements : np.ndarray(complex) or Tensor(complex)
        Motional displacements :math:`\{\eta_{jkl} \alpha_{jkl}\}` in phase-space with shape
        ``(3, N, N)`` without time samples or ``(T, 3, N, N)``, where ``T`` is the number
        of samples, ``3`` is the number of spatial axes, and ``N`` is the number of ions
        that is equal to the number of modes along an axis. The first dimension :math:`j` indicates
        the axis, the second dimension :math:`k` indicates the mode number along the axis,
        and the third dimension :math:`l` indicates the ion.
    target_phases : np.ndarray
        2D array containing target relative phases :math:`\{\psi_{kl}\}` between ion pairs.
        For ions :math:`k` and :math:`l`, with :math:`k > l`, the total relative phase target is
        the :math:`(k, l)`-th element.
        The `target_phases` must be a strictly lower triangular matrix.
    mean_phonon_numbers : np.ndarray or None, optional
        2D array with shape ``(3, N)`` of positive real numbers for each motional mode
        which corresponds to the mean phonon occupation :math:`\{\bar{n}_{jk}\}` of the given
        mode, where ``3`` is the number of spatial axes and ``N`` is the number of ions.
        If not provided, :math:`\bar{n}_{jk} = 0`, meaning no occupation of each mode.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor(real)
        A scalar or 1D tensor of infidelities with shape ``(T,)`` where ``T`` is
        the number of samples and one infidelity value per sample.

    See Also
    --------
    :func:`Graph.ms_dephasing_robust_cost <ions.ms_dephasing_robust_cost>`
        Cost for robust optimization of a Mølmer–Sørensen gate.
    :func:`Graph.ms_displacements <ions.ms_displacements>`
        Displacements for each mode/ion combination.
    :func:`Graph.ms_phases <ions.ms_phases>`
        Relative phases for all pairs of ions.

    Notes
    -----
    The infidelity is calculated according to [1]_

    .. math::
        1 - \mathcal{F}_\mathrm{av} = 1 - \left| \left( \prod_{\substack{k=1 \\ l<k}}^N \cos (
            \Phi_{kl} - \psi_{kl}) \right)
            \left( 1 - \sum_{j=1}^3 \sum_{k,l=1}^N \left[ |\eta_{jkl}|^2
            |\alpha_{jkl}|^2 \left(\bar{n}_{jk}+\frac{1}{2} \right) \right] \right) \right|^2 ,

    which assumes that the displacements :math:`\alpha_{jkl}` are
    small and eliminates terms of the fourth or higher order in them.

    References
    ----------
    .. [1] `C. D. B. Bentley, H. Ball, M. J. Biercuk, A. R. R. Carvalho,
            M. R. Hush, and H. J. Slatyer,
            Adv. Quantum Technol. 3, 2000044 (2020).
            <https://doi.org/10.1002/qute.202000044>`_

    Examples
    --------
    Refer to the `How to optimize error-robust Mølmer–Sørensen gates for trapped ions
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-error-robust-molmer-
    sorensen-gates-for-trapped-ions>`_ user guide to find how to use this and related nodes.
    """

    name = "ms_infidelity"
    args = [
        forge.arg("phases", type=TensorLike),
        forge.arg("displacements", type=TensorLike),
        forge.arg("target_phases", type=np.ndarray),
        forge.arg("mean_phonon_numbers", type=Optional[np.ndarray], default=None),
    ]
    rtype = Tensor
    categories = [Category.MOLMER_SORENSEN]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        phases = kwargs.get("phases")
        displacements = kwargs.get("displacements")
        target_phases = kwargs.get("target_phases")
        mean_phonon_numbers = kwargs.get("mean_phonon_numbers")
        check_numeric_numpy_array(phases, "phases")
        check_numeric_numpy_array(displacements, "displacements")
        check_numeric_numpy_array(target_phases, "target_phases")
        check_numeric_numpy_array(mean_phonon_numbers, "mean_phonon_numbers")
        phases_shape = validate_shape(phases, "phases")
        displacements_shape = validate_shape(displacements, "displacements")
        target_shape = validate_shape(target_phases, "target_phases")
        ion_count = target_shape[-1]
        check_argument(
            phases_shape[-2:] == (ion_count, ion_count),
            "The shape of phases must be (ion_count, ion_count) or"
            " (sample_count, ion_count, ion_count).",
            {"phases": phases},
            extras={"ion_count": ion_count, "phases.shape": phases_shape},
        )
        check_argument(
            len(phases_shape) <= 3,
            "The shape of phases must have at most 3 dimensions.",
            {"phases": phases},
            extras={"phases.shape": phases_shape},
        )
        check_argument(
            displacements_shape[-3:] == (3, ion_count, ion_count),
            "The shape of displacements must be (3, ion_count, ion_count) or"
            " (sample_count, 3, ion_count, ion_count).",
            {"displacements": displacements},
            extras={"ion_count": ion_count, "displacements.shape": displacements_shape},
        )
        check_argument(
            phases_shape[:-2] == displacements_shape[:-3],
            "If the shape of phases is (sample_count, ion_count, ion_count), then"
            " the shape of displacements must be (sample_count, 3, ion_count, ion_count).",
            {"phases": phases, "displacements": displacements},
            extras={
                "phases.shape": phases_shape,
                "displacements.shape": displacements_shape,
            },
        )
        check_argument(
            target_shape == (ion_count, ion_count),
            "The shape of target_phases must be (ion_count, ion_count).",
            {"target_phases": target_phases},
            extras={"ion_count": ion_count, "target_phases.shape": target_shape},
        )
        check_argument(
            np.allclose(target_phases, np.tril(target_phases, k=-1)),
            "The target_phases matrix must be strictly lower-triangular.",
            {"target_phases": target_phases},
        )
        if mean_phonon_numbers is not None:
            phonon_shape = validate_shape(mean_phonon_numbers, "mean_phonon_numbers")
            check_argument(
                phonon_shape == (3, ion_count),
                "The shape of mean_phonon_numbers must be (3, ion_count).",
                {"mean_phonon_numbers": mean_phonon_numbers},
                extras={
                    "ion_count": ion_count,
                    "mean_phonon_number.shape": phonon_shape,
                },
            )
        shape = phases_shape[:-2]
        return Tensor(_operation, shape=shape)


class MsDephasingRobustCost(Node):
    r"""
    Calculate the cost for robust optimization of a Mølmer–Sørensen gate.

    Add the tensor that this function returns to the infidelity of your
    target operation to obtain a cost that you can use to create a
    Mølmer–Sørensen gate that is robust against dephasing noise. You can
    further multiply the robust cost by a scaling factor to weigh how much
    importance you give to the robustness compared to the original cost.

    Parameters
    ----------
    drives : list[Pwc(1D, complex)]
        The list of piecewise-constant drives :math:`\{\gamma_j\}`. The
        number of drives must be the same as the number of ions :math:`N`.
        Drive values must be one-dimensional arrays and in rad/s. Drive durations must be in
        seconds. All drives must have the same total duration, but they can
        have different numbers of segments.
    lamb_dicke_parameters : np.ndarray
        A ``(3, N, N)`` :math:`\{ \eta_{jkn} \}` array of parameters
        specifying the laser-ion coupling strength, where :math:`N` equals
        the number of ions. The first dimension :math:`j` indicates the
        axis, the second dimension :math:`k` indicates the collective mode
        number, and the third dimension :math:`n` indicates the ion.
    relative_detunings : np.ndarray
        The 2D array :math:`\{\delta_{jk} = \nu_{jk} - \delta\}` specifying
        the difference (in Hz) between each motional mode frequency
        :math:`\nu_{jk}` and the laser detuning :math:`\delta` (the
        detuning from the qubit transition frequency). Its shape must be
        ``(3, N)`` where the first dimension :math:`j` indicates the axis
        and the second dimension :math:`k` indicates the collective mode
        number.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor(scalar, real)
        The cost term that you can use to optimize a Mølmer–Sørensen gate
        that is robust against dephasing noise. The cost is the sum of the
        square moduli of the time-averaged positions of the phase-space
        trajectories, weighted by the corresponding Lamb–Dicke parameters.

    See Also
    --------
    :func:`Graph.ions.ms_infidelity <ions.ms_infidelity>`
        Final operational infidelity of a Mølmer–Sørensen gate.
    :func:`boulderopal.ions.obtain_ion_chain_properties`

    Notes
    -----
    You can construct a Mølmer–Sørensen gate that is robust against
    dephasing noise by a combination of minimizing the time-averaged
    positions of the phase-space trajectories and imposing a symmetry in
    each ion's drive [1]_.

    The displacement of the :math:`j`-th ion in the :math:`p`-th mode of
    oscillation is the following [2]_:

    .. math::
        \alpha_{pj}(t) = \int_0^t d\tau \frac{\gamma_j(\tau)}{2}
        e^{i \delta_p \tau} .

    where the axis dimension and the collective mode dimension are combined
    into a single index :math:`p` for simplicity. For a gate of duration
    :math:`t_\text{gate}`, the time-averaged position is:

    .. math::
        \langle \alpha_{pj} \rangle = \frac{1}{t_\text{gate}}
        \int_0^{t_\text{gate}} \alpha_{pj}(t) \mathrm{d} t .

    This function returns the sum of the square moduli of the time-averaged
    positions multiplied by the corresponding Lamb–Dicke parameters. These
    parameters weight the time-averaged positions in the same way that the
    :math:`\alpha_{pj}(t)` are weighted in the formula for the infidelity
    of a Mølmer–Sørensen gate.

    In other words, the robust cost that this function returns is:

    .. math::
        C_\text{robust} = \sum_{p,j} \left| \eta_{pj} \langle \alpha_{pj}
        \rangle \right|^2.

    You can add this to the infidelity with the respect to the target gate
    to create the cost function that optimizes a gate that is also robust
    against dephasing. You can further multiply :math:`C_\text{robust}` by
    a scaling factor to weigh how much importance you give to robustness.

    References
    ----------
    .. [1] `A. R. Milne, C. L. Edmunds, C. Hempel, F. Roy, S. Mavadia,
            and M. J. Biercuk,
            Phys. Rev. Appl. 13, 024022 (2020).
            <https://doi.org/10.1103/PhysRevApplied.13.024022>`_
    .. [2] `C. D. B. Bentley, H. Ball, M. J. Biercuk, A. R. R. Carvalho,
            M. R. Hush, and H. J. Slatyer,
            Adv. Quantum Technol. 3, 2000044 (2020).
            <https://doi.org/10.1002/qute.202000044>`_

    Examples
    --------
    Refer to the `How to optimize error-robust Mølmer–Sørensen gates for trapped ions
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-error-robust-molmer-
    sorensen-gates-for-trapped-ions>`_ user guide to find how to use this and related nodes.
    """

    name = "ms_dephasing_robust_cost"
    args = [
        forge.arg("drives", type=List[Pwc]),
        forge.arg("lamb_dicke_parameters", type=np.ndarray),
        forge.arg("relative_detunings", type=np.ndarray),
    ]
    rtype = Tensor
    categories = [Category.MOLMER_SORENSEN]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        drives = kwargs.get("drives")
        lamb_dicke_parameters = kwargs.get("lamb_dicke_parameters")
        relative_detunings = kwargs.get("relative_detunings")
        check_argument_iterable(drives, "drives")
        check_argument(
            len(drives) > 0, "At least one drive must be provided.", {"drives": drives}
        )
        check_argument(
            all(isinstance(drive, Pwc) for drive in drives),
            "Each of the drives must be a Pwc.",
            {"drives": drives},
        )
        check_argument(
            all(len(drive.values.shape) == 1 for drive in drives),
            "The value of each drive must be 1D.",
            {"drives": drives},
        )
        ion_count = len(drives)
        validate_ms_shapes(
            ion_count=ion_count,
            ld_values=lamb_dicke_parameters,
            ld_name="lamb_dicke_parameters",
            rd_values=relative_detunings,
            rd_name="relative_detunings",
        )
        return Tensor(_operation, shape=())


class IonsNamespace:
    """
    Operations to describeMølmer–Sørensen interactions.
    """

    __name__ = "ions"

    def __init__(self, graph):
        self._graph = graph

    # pylint: disable=missing-function-docstring
    @wraps(MsDephasingRobustCost.create_graph_method())
    def ms_dephasing_robust_cost(self, *args, **kwargs):
        return MsDephasingRobustCost.create_graph_method()(self._graph, *args, **kwargs)

    @wraps(MsDisplacements.create_graph_method())
    def ms_displacements(self, *args, **kwargs):
        return MsDisplacements.create_graph_method()(self._graph, *args, **kwargs)

    @wraps(MsInfidelity.create_graph_method())
    def ms_infidelity(self, *args, **kwargs):
        return MsInfidelity.create_graph_method()(self._graph, *args, **kwargs)

    @wraps(MsPhases.create_graph_method())
    def ms_phases(self, *args, **kwargs):
        return MsPhases.create_graph_method()(self._graph, *args, **kwargs)

    @wraps(MsPhasesMultitone.create_graph_method())
    def ms_phases_multitone(self, **kwargs):
        return MsPhasesMultitone.create_graph_method()(self._graph, **kwargs)


IonsNamespaceDoc = create_namespace_doc("ions", __name__, [Category.MOLMER_SORENSEN])



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

# Make private alias for all imports to clean up the autocomplete for the graph module.
import sys as _sys

from boulderopal._nodes.composite.namespace import NAMESPACE_ATTR as _ns
from boulderopal._nodes.composite.registry import (
    COMPOSITE_NAMESPACE_REGISTRY as _ns_registry,
)
from boulderopal._nodes.registry import (
    PRIMARY_NAMESPACE_DOCS as _primary_namespace_docs,
)
from boulderopal._nodes.registry import TYPE_REGISTRY as _type_registry
from boulderopal.graph._execute_graph import (
    ExecutionMode,
    execute_graph,
)
from boulderopal.graph._graph import Graph

# Binding the Graph related types to the current module, this is useful
# for building docs for the Graph class.
_module = _sys.modules[__name__]
for _type_cls in _type_registry:
    setattr(_module, _type_cls.__name__, _type_cls)


for _primary_namespace in _primary_namespace_docs:
    # Create namespace.
    setattr(_module, _primary_namespace.namespace_name, _primary_namespace)

    # Bind namespace methods.
    for _node_name, _node_method in _primary_namespace.sub_nodes.items():
        setattr(
            getattr(_module, _primary_namespace.namespace_name),
            _node_name,
            _node_method,
        )


# Set composite namespaces to Graph.
# This import-time binding is only useful for building the docs.
# The actual composite namespace object is created during the
# Graph object initialization time.
for _composite_namespace in _ns_registry:
    setattr(_module, getattr(_composite_namespace, _ns), _composite_namespace)


__all__ = ["ExecutionMode", "Graph", "execute_graph"]



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from collections import deque
from itertools import chain
from typing import (
    TYPE_CHECKING,
    Iterable,
)

from qctrlcommons.exceptions import QctrlArgumentsValueError
from qctrlcommons.node.wrapper import (
    NodeData,
    Operation,
)
from qctrlcommons.preconditions import check_argument

from boulderopal._nodes.registry import PRIMARY_NODE_REGISTRY

if TYPE_CHECKING:
    from boulderopal.graph import Graph


def validate_output_node_names(
    node_names: str | list[str], graph: "Graph"
) -> list[str]:
    """
    Validate the names of the output nodes for fetching from a graph.

    If any node is not in the graph, raise an error. Otherwise, normalize the names to a list of
    strings.

    Parameters
    ---------
    node_names : str or list[str]
        Name of the nodes to be fetched.
    graph : Graph
        The graph where the nodes are supposed to be fetched from.

    Returns
    -------
    list[str]
        A list of valid node names.
    """

    if isinstance(node_names, str):
        node_names = [node_names]

    check_argument(
        isinstance(node_names, list)
        and all(isinstance(name, str) for name in node_names),
        "The output node names must be a string or a list of strings.",
        {"output_node_names": node_names},
    )

    check_argument(
        len(node_names) >= 1,
        "The output node names must have at least one element.",
        {"output_node_names": node_names},
    )

    for name in node_names:
        check_node_in_graph(
            name,
            graph,
            f"The requested output node name '{name}' is not present in the graph.",
        )

    return node_names


def check_node_in_graph(node: str, graph: "Graph", message: str) -> None:
    """
    Check if a node is in the Graph.

    Parameters
    ----------
    node : str
        The name of the node.
    graph : Graph
        The Graph to be validated.
    message : str
        The error message.
    """
    check_argument(
        node in graph.operations, message, {"graph": graph}, extras={"node name": node}
    )


def check_optimization_node_in_graph(graph: "Graph") -> None:
    """
    Check optimization graph at least has one optimization node.
    """
    for operation in graph.operations.values():
        if PRIMARY_NODE_REGISTRY.get_node_cls(
            operation.operation_name
        ).optimizable_variable:
            return
    raise QctrlArgumentsValueError(
        "At least one optimization variable is required in the optimization graph.",
        {"graph": graph},
    )


def check_cost_node(node: str, graph: Graph) -> None:
    """
    Check cost node:
        - if the node is in the graph.
        - if the node is a scalar Tensor.
    """
    check_argument(
        isinstance(node, str),
        "The cost node name must be a string.",
        {"cost_node_name": node},
    )
    check_node_in_graph(node, graph, "A cost node must be present in the graph.")
    check_argument(
        graph.operations[node].is_scalar_tensor,
        "The cost node must be a scalar Tensor.",
        {"cost": node},
        extras={"cost node operation": graph.operations[node]},
    )


def check_cost_node_for_optimization_graph(
    graph: Graph,
    cost_node_name: str,
    output_node_names: list[str],
    check_gradient_nodes: bool = True,
) -> None:
    """
    Traverse the graph from the cost node, and check:
        1. All connected the nodes should support gradient if `check_gradient_nodes` is True.
        2. Any optimizable node to be fetched should connect to the cost node.
    """

    connected_optimization_node_names = set()

    def _validate_node_from_operation(operation: Operation) -> None:
        node = PRIMARY_NODE_REGISTRY.get_node_cls(operation.operation_name)
        if check_gradient_nodes:
            check_argument(
                node.supports_gradient,
                f"The {operation.operation_name} node does not support gradient.",
                {"graph": graph},
            )
        if node.optimizable_variable:
            connected_optimization_node_names.add(operation.name)

    def _get_parent_operations(node: str) -> Iterable[Operation]:
        """
        Go through inputs of the nodes, which might include Python primitive iterables.
        Find all NodeData and flat them as a single iterable.
        """

        def _get_input_items(input_: Iterable) -> Iterable:
            if isinstance(input_, NodeData):
                return [input_.operation]
            if isinstance(input_, (list, tuple)):
                return chain.from_iterable(_get_input_items(item) for item in input_)
            if isinstance(input_, dict):
                return chain.from_iterable(
                    _get_input_items(item) for item in input_.values()
                )
            return []

        return chain.from_iterable(
            _get_input_items(input_)
            for input_ in graph.operations[node].kwargs.values()
        )

    visited_nodes: set[str] = set()
    nodes_to_check: deque = deque()

    # cost node is where we start with.
    _validate_node_from_operation(graph.operations[cost_node_name])
    visited_nodes.add(cost_node_name)
    nodes_to_check.appendleft(cost_node_name)

    while nodes_to_check:
        node = nodes_to_check.pop()

        for operation in _get_parent_operations(node):
            if operation.name not in visited_nodes:
                _validate_node_from_operation(operation)
                visited_nodes.add(operation.name)
                nodes_to_check.appendleft(operation.name)

    # Graph traverse is done and all connected optimization nodes are recorded.
    # Now check output nodes.
    for output_node in output_node_names:
        op_name = graph.operations[output_node].operation_name
        if PRIMARY_NODE_REGISTRY.get_node_cls(op_name).optimizable_variable:
            check_argument(
                output_node in connected_optimization_node_names,
                "The requested optimization node in `output_node_names` is not connected "
                "to the cost node.",
                {"output_node_names": output_node_names},
                extras={"disconnected output node name": output_node},
            )


def check_initial_value_for_optimization_node(graph: Graph) -> None:
    """
    Check optimization node has valid non-default initial values.
    """

    initial_value_info = {}

    for name, operation in graph.operations.items():
        node = PRIMARY_NODE_REGISTRY.get_node_cls(operation.operation_name)
        if (
            node.optimizable_variable
            and operation.kwargs.get("initial_values") is not None
        ):
            initial_value_info[name] = operation.kwargs["initial_values"]

    initial_values = list(initial_value_info.values())
    if len(initial_values) != 0:
        for val in initial_values[1:]:
            if not isinstance(val, type(initial_values[0])):
                raise QctrlArgumentsValueError(
                    "Non-default initial values of optimization variables in the graph"
                    " must all either be an array or a list of arrays.",
                    {"graph": graph},
                    extras=initial_value_info,
                )

        if isinstance(initial_values[0], list):
            for val in initial_values[1:]:
                if len(val) != len(initial_values[0]):
                    raise QctrlArgumentsValueError(
                        "Lists of initial values of optimization variables must have "
                        "the same length.",
                        {"graph": graph},
                        extras=initial_value_info,
                    )
